searchNodes=[{"doc":"This allows items extracted from Kinesis stream records (or sub-records in a KPL aggregate record ) to be processed by a pipeline of workers which may differ in number from the number of shards owned by the current node (which is the normal processing model offered by erlmld ). This is beneficial when using aggregate records which can be processed in approximate order according to their partition keys as opposed to strict ordering based on the shards they arrived on. For example, suppose the following two Kinesis records are received on two different shards: Record 1 ( a KPL aggregate record ) - partition key : &quot;xyzzy&quot; - subrecord a : - partition key : &quot;asdf&quot; - value : &quot;12345&quot; - subrecord b : - partition key : &quot;fdsa&quot; - value : &quot;54321&quot; Record 2 ( a KPL aggregate record ) - partition key : &quot;qwer&quot; - subrecord a : - partition key : &quot;asdf&quot; - value : &quot;23456&quot; - subrecord b : - partition key : &quot;z&quot; - value : &quot;0&quot; Using the normal Kinesis processing paradigm, each shard will be processed in order. erlmld supports this by spawning a process for each owned shard, which handles each record seen on the shard in sequence: Worker 1 : 1 . handle record &quot;xyzzy&quot; a . handle sub - record &quot;asdf&quot; b . handle sub - record &quot;fdsa&quot; Worker 2 : 1 . handle record &quot;qwer&quot; a . handle sub - record &quot;asdf&quot; b . handle sub - record &quot;z&quot; This can fail to make use of all available resources since the maximum concurrency is limited by the number of owned shards. If the application can tolerate the handling of sub-records in a non-strict order, it can use a Flow -based MapReduce-style scheme: [ Worker 1 ] [ Worker 2 ] ( processes which produce Kinesis records ) | | v v [ Exmld.KinesisStage , ... ] ( stages receiving Exmld.KinesisWorker.Datums ) | v [ M1 ] ... . [ Mn ] ( mappers which extract items ) | \\ / | | \\ / | | \\ / | | \\ / | | \\ | | / \\ | | / \\ | | / \\ | | / \\ | [ R1 ] ... . [ Rn ] ( reducers which handle extracted items ) The number of reducers is configurable and defaults to the number of schedulers online. The processing application will specify a means of extracting a partition key from each extracted item; these will be used to consistently map items to reducers (which is where the actual application work occurs). Using the above example and specifying a sub-record's partition key as an item key: Worker 1 will produce the &quot;asdf&quot; and &quot;fdsa&quot; sub-records from outer record &quot;xyzzy&quot; and send them to a pre-configured Exmld.KinesisStage (or round-robin to a list of such stages). Worker 2 will similarly produce the &quot;asdf&quot; and &quot;z&quot; sub-records from outer record &quot;qwer&quot;. Each receiving stage will wrap and forward these sub-records for handling by the flow. The application will have provided an &quot;identity&quot; item extraction function since KPL aggregation is being used here (or otherwise a function accepting one record and returning a list containing a single item). The application will have provided a partition key extraction function which returns an appropriate partition key to be used in consistently mapping items to reducers. The first received &quot;asdf&quot; sub-record is provided to some reducer Rx . The second received &quot;asdf&quot; sub-record is provided to the same reducer since its extracted key has the same hash. The &quot;fdsa&quot; and &quot;z&quot; sub-records are similarly provided to some worker Ry and/or Rz based on the hash of their partition keys. The application-provided reducer function notifies each originating stage of the disposition of processing for items received from it as processing progresses. Eventually, processing disposition is provided back to the originating workers, which can decide whether or not (and where) to checkpoint.","ref":"Exmld.html","title":"Exmld","type":"module"},{"doc":"","ref":"Exmld.html#checkpoint/1","title":"Exmld.checkpoint/1","type":"macro"},{"doc":"","ref":"Exmld.html#checkpoint/2","title":"Exmld.checkpoint/2","type":"macro"},{"doc":"","ref":"Exmld.html#child_spec/1","title":"Exmld.child_spec/1","type":"function"},{"doc":"Accepts a flow producing Exmld.KinesisWorker.Datum s (e.g,. a flow created from Exmld.KinesisStage s) and returns another flow.","ref":"Exmld.html#flow/6","title":"Exmld.flow/6","type":"function"},{"doc":"You can use this one to keep building your flow after calling flow/6 above.","ref":"Exmld.html#from_stages/1","title":"Exmld.from_stages/1","type":"function"},{"doc":"","ref":"Exmld.html#sequence_number/1","title":"Exmld.sequence_number/1","type":"macro"},{"doc":"","ref":"Exmld.html#sequence_number/2","title":"Exmld.sequence_number/2","type":"macro"},{"doc":"","ref":"Exmld.html#start_link/1","title":"Exmld.start_link/1","type":"function"},{"doc":"","ref":"Exmld.html#stream_record/1","title":"Exmld.stream_record/1","type":"macro"},{"doc":"","ref":"Exmld.html#stream_record/2","title":"Exmld.stream_record/2","type":"macro"},{"doc":"","ref":"Exmld.html#t:checkpoint/0","title":"Exmld.checkpoint/0","type":"type"},{"doc":"","ref":"Exmld.html#t:item/0","title":"Exmld.item/0","type":"type"},{"doc":"","ref":"Exmld.html#t:partition_key/0","title":"Exmld.partition_key/0","type":"type"},{"doc":"","ref":"Exmld.html#t:reducer_state/0","title":"Exmld.reducer_state/0","type":"type"},{"doc":"","ref":"Exmld.html#t:sequence_number/0","title":"Exmld.sequence_number/0","type":"type"},{"doc":"","ref":"Exmld.html#t:shard_id/0","title":"Exmld.shard_id/0","type":"type"},{"doc":"","ref":"Exmld.html#t:stream_record/0","title":"Exmld.stream_record/0","type":"type"},{"doc":"A GenStage stage for use in processing data produced by Exmld.KinesisWorker s. This module acts as a GenStage producer. Subscribers will receive Exmld.KinesisStage.Event s, which each wrap an underlying Exmld.KinesisWorker.Datum along with information about the producing worker and stage. Downstream processors should eventually call disposition/2 with the disposition of processing so that originating workers can checkpoint. The expected use and workflow is: Create a stage or set of stages using this module for each distinct processing pipeline. A set of stages could be used by more than one Kinesis stream if the processing being done is the same for all of them. Create a flow using Flow.from_stages/2 . Configure the flow using Exmld.flow/6 . Run the flow, which should run forever. Configure an erlmld supervision tree with a set of Exmld.KinesisWorker s using the stage(s) created in (1).","ref":"Exmld.KinesisStage.html","title":"Exmld.KinesisStage","type":"module"},{"doc":"Notify stage of the disposition of processing some items. An attempt has been made to process some data extracted from a Kinesis record by a downstream processor. stage will look up the originating producer and record the disposition of processing in the next batch of data to be returned to that producer.","ref":"Exmld.KinesisStage.html#disposition/3","title":"Exmld.KinesisStage.disposition/3","type":"function"},{"doc":"Callback implementation for GenStage.handle_call/3 .","ref":"Exmld.KinesisStage.html#handle_call/3","title":"Exmld.KinesisStage.handle_call/3","type":"function"},{"doc":"Handle subscriber demand. Return up to incoming_demand + pending_demand events, fetching (from state) as needed, and storing in state any excess. If not enough events are available, record unsatisfied demand in state, and then return those events when answering a subsequent call. See the QueueBroadcaster example in GenStage for an explanation of this demand queueing behavior.","ref":"Exmld.KinesisStage.html#handle_demand/2","title":"Exmld.KinesisStage.handle_demand/2","type":"function"},{"doc":"Callback implementation for GenStage.handle_info/2 .","ref":"Exmld.KinesisStage.html#handle_info/2","title":"Exmld.KinesisStage.handle_info/2","type":"function"},{"doc":"Callback implementation for GenStage.init/1 .","ref":"Exmld.KinesisStage.html#init/1","title":"Exmld.KinesisStage.init/1","type":"function"},{"doc":"Notify stage of a new Kinesis record available for processing. A new event is available for processing by stage . The caller will be monitored and associated with the new event, and will be blocked until after the event has been used to satisfy some downstream demand. The return value will be the disposition (success/failure) of zero or more records which were previously processed.","ref":"Exmld.KinesisStage.html#notify/3","title":"Exmld.KinesisStage.notify/3","type":"function"},{"doc":"","ref":"Exmld.KinesisStage.html#start_link/1","title":"Exmld.KinesisStage.start_link/1","type":"function"},{"doc":"","ref":"Exmld.KinesisStage.Event.html","title":"Exmld.KinesisStage.Event","type":"module"},{"doc":"Struct for events provided to an Exmld.KinesisStage . Records the stage and worker identifiers associated with an event. Fields :stage - identifier of the Exmld.KinesisStage which handled the event :worker - identifier of the Exmld.KinesisWorker which produced the event :event - an Exmld.KinesisWorker.Datum","ref":"Exmld.KinesisStage.Event.html#__struct__/0","title":"Exmld.KinesisStage.Event.__struct__/0","type":"function"},{"doc":"","ref":"Exmld.KinesisStage.Event.html#t:t/0","title":"Exmld.KinesisStage.Event.t/0","type":"type"},{"doc":"An erlmld_flusher which can interface with a Exmld.KinesisStage data source. This implements an erlmld_flusher which can be used by erlmld_batch_processor . Unlike a typical erlmld_flusher , it has a different notion of fullness: if more than :max_pending items are in flight, the worker waits for all pending items before emitting any more for downstream processing. A periodic flush interval should be configured in the batch processor options. Similarly, the downstream stage processing pipeline should not require any kind of &quot;full&quot; condition and should periodically make progress (i.e., emit/flush output) even if no more records are sent. Heartbeat items are sent while the worker is waiting for pending items to be completed; these include varying counters to allow them to be automatically distributed among downstream reducers. One worker process will exist for each stream shard owned by the current node. Each such process will have been configured with a set of downstream Exmld.KinesisStage s which can receive records from it (actually Exmld.KinesisWorker.Datum s); those stages will be part of a data processing Flow.t . Eventually, the disposition of each record's processing will propagate back to the originating worker (as return values from GenStage.call/3 ). Periodically, erlmld_batch_processor will request a flush. If the flush kind is :partial , we return the tokens associated with the records which have already been fully processed. Otherwise, the flush kind is :full and we await the disposition of every outstanding record before returning. If processing of any record (or item extracted therefrom) fails, the worker will crash unless it's configured to ignore processing errors. Records presented to this worker may be ordinary records or sub-records extracted from a containing KPL-aggregated record. If KPL aggregation is not being used, but smaller sub-items are later extracted by the stage processing pipeline, the pipeline should create fake sub-record sequence numbers to track the disposition of those items (and sub-record checkpointing should be turned off). Periodically (which should be at some multiple of the periodic flush interval), erlmld_batch_processor will checkpoint based on the records which have so far been successfully processed (those whose tokens have been returned from flush/2 ).","ref":"Exmld.KinesisWorker.html","title":"Exmld.KinesisWorker","type":"module"},{"doc":"Submit a new Kinesis record to the downstream pipeline for processing. A new Kinesis record is available for processing, and erlmld_batch_processor is instructing us to add it to the current batch. Since we really have no notion of a batch, we immediately choose a downstream stage and notify it of a new Exmld.KinesisWorker.Datum containing the record and make a note of it being in-flight. That call will block until a further-downstream consumer receives the record as a flow event. The result of that call will be an updated list of item dispositions. Unless configured to skip records which failed to be processed, we crash if any failed. Otherwise we update the set of done/pending items and return an updated state.","ref":"Exmld.KinesisWorker.html#add_record/3","title":"Exmld.KinesisWorker.add_record/3","type":"function"},{"doc":"Return a list of tokens corresponding to records which have been fully processed and the latest state. If the flush kind is :full , we await the disposition of all outstanding records before returning. Otherwise, it's :partial and we return (possibly an empty result) immediately. If doing a full flush and any records fail to be successfully processed, we crash unless configured to skip failed records.","ref":"Exmld.KinesisWorker.html#flush/2","title":"Exmld.KinesisWorker.flush/2","type":"function"},{"doc":"The batch processor has received a possibly-empty set of records from the MultiLangDaemon and is informing the flusher (us). Send a heartbeat downstream and return any completed tokens. This allows progress to be made even if no more records appear on the stream.","ref":"Exmld.KinesisWorker.html#heartbeat/1","title":"Exmld.KinesisWorker.heartbeat/1","type":"function"},{"doc":"Initialize worker state with a shard id and a set of options. An erlmld_batch_processor is initializing processing on shard_id and providing the flusher_mod_data which was passed to it, which should be an enumerable of keyword s containing the following options; we return a flusher state to be used in subsequent operations. Options All optional unless marked required: :stages - (required) list of GenStage s (values useable as first arg to GenStage.call/3 ) which can receive Exmld.KinesisWorker.Datum s :opaque - opaque term passed in each Exmld.KinesisWorker.Datum :skip_errors - boolean indicating whether errors are non-fatal (if false, crash on error). :max_pending - maximum number of pending items which can be in flight. :await_sleep_interval - sleep time between checks while awaiting pending items. :error_callback - nil or an arity-2 function called with state and failure dispositions when processing failures occur.","ref":"Exmld.KinesisWorker.html#init/2","title":"Exmld.KinesisWorker.init/2","type":"function"},{"doc":"","ref":"Exmld.KinesisWorker.html#t:flusher_token/0","title":"Exmld.KinesisWorker.flusher_token/0","type":"type"},{"doc":"","ref":"Exmld.KinesisWorker.html#t:t/0","title":"Exmld.KinesisWorker.t/0","type":"type"},{"doc":"","ref":"Exmld.KinesisWorker.Datum.html","title":"Exmld.KinesisWorker.Datum","type":"module"},{"doc":"Struct for annotating stream records (or heartbeats) with additional data. To allow a single downstream processing pipeline to be used with multiple source streams, we annotate Kinesis stream records with additional data before providing them to the stage(s) processing them. (Example: :opaque could be a 2-tuple naming the source stream and region or otherwise indicate how to specially process the record). Fields :opaque - the opaque term provided at worker init time :shard_id - name of the shard the worker is processing :stream_record - a record from the stream or {:heartbeat, _}","ref":"Exmld.KinesisWorker.Datum.html#__struct__/0","title":"Exmld.KinesisWorker.Datum.__struct__/0","type":"function"},{"doc":"","ref":"Exmld.KinesisWorker.Datum.html#t:t/0","title":"Exmld.KinesisWorker.Datum.t/0","type":"type"},{"doc":"","ref":"Exmld.KinesisWorker.Disposition.html","title":"Exmld.KinesisWorker.Disposition","type":"module"},{"doc":"Struct for event processing disposition. Tracks whether processing succeeded or failed for a specific record or item extracted therefrom. Fields :sequence_number - Exmld.sequence_number() of the subject record. If the subject is an item extracted from a containing aggregate record, the user_sub and user_total fields should be populated (whether KPL aggregation was used or not). :status - processing status","ref":"Exmld.KinesisWorker.Disposition.html#__struct__/0","title":"Exmld.KinesisWorker.Disposition.__struct__/0","type":"function"},{"doc":"","ref":"Exmld.KinesisWorker.Disposition.html#t:t/0","title":"Exmld.KinesisWorker.Disposition.t/0","type":"type"}]